import k_ctds as ctds

from .base import TestExternalDatabase

class TestCursorFetchOne(TestExternalDatabase):
    '''Unit tests related to the Cursor.fetchone() method.
    '''
    def test___doc__(self):
        self.assertEqual(
            ctds.Cursor.fetchone.__doc__,
            '''\
fetchone()

Fetch the next row of a query result set, returning a single sequence, or
:py:data:`None` when no more data is available.

:pep:`0249#fetchone`

:return: The next row or :py:data:`None`.
'''
        )

    def test_closed(self):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.close()
            try:
                cursor.fetchone()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'cursor closed')
            else:
                self.fail('.fetchone() did not fail as expected') # pragma: nocover

    def test_closed_connection(self):
        connection = self.connect()
        with connection.cursor() as cursor:
            connection.close()
            try:
                cursor.fetchone()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'connection closed')
            else:
                self.fail('.fetchone() did not fail as expected') # pragma: nocover

    def test_premature(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_fetchone(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT * FROM @{0};
                        SELECT i * 2 FROM @{0};
                    '''.format(self.test_fetchone.__name__)
                )
                self.assertEqual(tuple(cursor.fetchone()), (1,))
                self.assertEqual(tuple(cursor.fetchone()), (2,))
                self.assertEqual(tuple(cursor.fetchone()), (3,))
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual(tuple(cursor.fetchone()), (2,))
                self.assertEqual(tuple(cursor.fetchone()), (4,))
                self.assertEqual(tuple(cursor.fetchone()), (6,))
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), None)
                self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_empty_resultset(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT i FROM @{0} WHERE i < 0;
                    '''.format(self.test_empty_resultset.__name__)
                )
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), None)

    def test_multiple_resultsets(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT i FROM @{0} WHERE i < 0;
                        SELECT i AS j FROM @{0} WHERE i > 2;
                        SELECT i AS k FROM @{0} WHERE i > 3;
                        SELECT i AS ii FROM @{0};
                    '''.format(self.test_multiple_resultsets.__name__)
                )
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual(tuple(cursor.fetchone()), (3,))
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual(tuple(cursor.fetchone()), (1,))
                self.assertEqual(tuple(cursor.fetchone()), (2,))
                self.assertEqual(tuple(cursor.fetchone()), (3,))
                self.assertEqual(cursor.nextset(), None)



import k_ctds as ctds

from .base import TestExternalDatabase

class TestCursorEnter(TestExternalDatabase):

    def test___doc__(self):
        self.assertEqual(
            ctds.Cursor.__enter__.__doc__,
            '''\
__enter__()

Enter the cursor's runtime context. On exit, the cursor is
closed automatically.

:return: The cursor object.
:rtype: k_ctds.Cursor
'''
        )



from datetime import date, datetime, time
from binascii import hexlify, unhexlify
from decimal import Decimal
import platform
import uuid

import k_ctds as ctds

from .base import TestExternalDatabase
from .compat import unichr_, unicode_


class TestSQLToPython(TestExternalDatabase): # pylint: disable=too-many-public-methods
    '''Unit tests related to SQL to Python type conversion.
    '''

    def setUp(self):
        TestExternalDatabase.setUp(self)

        self.connection = self.connect()
        self.cursor = self.connection.cursor()

    def tearDown(self):
        TestExternalDatabase.tearDown(self)

        self.cursor.close()
        self.connection.close()

    def test_notsupportederror(self):
        self.cursor.execute("SELECT sql_variant_property(1, 'BaseType')")
        try:
            self.cursor.fetchone()
        except ctds.NotSupportedError as ex:
            self.assertEqual(str(ex), 'unsupported type 98 for column ""')
        else:
            self.fail('.fetchone() did not fail as expected') # pragma: nocover

    def test_bit(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(BIT, NULL),
                CONVERT(BIT, 1),
                CONVERT(BIT, 0)
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (None, True, False)
        )

    def test_binary(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(BINARY(8), NULL),
                CONVERT(BINARY(8), 0xdeadbeef),

                CONVERT(VARBINARY(8), NULL),
                CONVERT(VARBINARY(16), 0xdeadbeef)
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                unhexlify('deadbeef00000000'),

                None,
                unhexlify('deadbeef'),
            )
        )

    def test_char(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(CHAR(4), NULL),
                CONVERT(CHAR(4), '1234'),

                CONVERT(VARCHAR(4), NULL),
                CONVERT(VARCHAR(4), '1234'),
                CONVERT(VARCHAR(4), ''),
                CONVERT(VARCHAR(4), '1'),
                REPLICATE(CONVERT(VARCHAR(MAX), 'x'), 8001),

                CONVERT(CHAR(1), CHAR(189))
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                unicode_('1234'),
                None,
                unicode_('1234'),
                unicode_(''),
                unicode_('1'),
                unicode_('x' * 8001),
                unicode_(b'\xc2\xbd', encoding='utf-8'),
            )
        )

    def test_nchar(self):
        non_ucs2_emoji = unichr_(127802) if self.UCS4_SUPPORTED else self.UNICODE_REPLACEMENT
        self.cursor.execute(
            '''
            SELECT
                CONVERT(NCHAR(4), NULL),
                CONVERT(NCHAR(4), '1234'),

                CONVERT(NVARCHAR(4), NULL),
                CONVERT(NVARCHAR(4), N'1234'),
                CONVERT(NVARCHAR(4), N''),
                CONVERT(NVARCHAR(4), N'1'),
                REPLICATE(CONVERT(NVARCHAR(MAX), N'x'), 8001),

                NCHAR(189),
                NCHAR(256),
                CONVERT(NVARCHAR(100), 0x{0})
            '''.format(hexlify(non_ucs2_emoji.encode('utf-16le')).decode('ascii'))
        )

        # Windows builds properly decode codepoints from the supplementary plane.
        # Possibly due to the iconv implementation??
        decoded = self.use_utf16 or platform.system() == 'Windows' or not self.UCS4_SUPPORTED
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                unicode_('1234'),
                None,
                unicode_('1234'),
                unicode_(''),
                unicode_('1'),
                unicode_('x' * 8001),
                unicode_(b'\xc2\xbd', encoding='utf-8'),
                unicode_(b'\xc4\x80', encoding='utf-8'),
                non_ucs2_emoji if decoded else unicode_('??')
            )
        )

    def test_text(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(TEXT, NULL),
                CONVERT(TEXT, N''),
                CONVERT(TEXT, N'1234')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                unicode_(''),
                unicode_('1234')
            )
        )

    def test_ntext(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(NTEXT, NULL),
                CONVERT(NTEXT, N''),
                CONVERT(NTEXT, N'1234')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                unicode_(''),
                unicode_('1234')
            )
        )

    def test_int(self):
        ints = (2 ** 8 - 1, 2 ** 15 - 1, 2 ** 31 - 1, 2 ** 63 - 1)
        self.cursor.execute(
            '''
            SELECT
                CONVERT(TINYINT, NULL),
                CONVERT(TINYINT, {0}),

                CONVERT(SMALLINT, NULL),
                CONVERT(SMALLINT, {1}),
                CONVERT(SMALLINT, -{1}),

                CONVERT(INT, NULL),
                CONVERT(INT, {2}),
                CONVERT(INT, -{2}),

                CONVERT(BIGINT, NULL),
                CONVERT(BIGINT, {3}),
                CONVERT(BIGINT, -{3})
            '''.format(*ints)
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None, ints[0],
                None, ints[1], -ints[1],
                None, ints[2], -ints[2],
                None, ints[3], -ints[3],
            )
        )

    def test_float(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(REAL, NULL),
                CONVERT(REAL, '-3.40E+38'),
                CONVERT(REAL, '-1.18E-38'),
                CONVERT(REAL, 0),
                CONVERT(REAL, '1.18E-38'),
                CONVERT(REAL, '3.40E+38'),

                CONVERT(FLOAT(24), NULL),
                CONVERT(FLOAT(24), '-3.40E+38'),
                CONVERT(FLOAT(24), '-1.18E-38'),
                CONVERT(FLOAT(24), 0),
                CONVERT(FLOAT(24), '1.18E-38'),
                CONVERT(FLOAT(24), '3.40E+38'),

                CONVERT(FLOAT(53), NULL),
                CONVERT(FLOAT(53), '-1.79E+308'),
                CONVERT(FLOAT(53), '-2.23E-308'),
                CONVERT(FLOAT(53), 0),
                CONVERT(FLOAT(53), '2.23E-308'),
                CONVERT(FLOAT(53), '1.79E+308')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                -3.3999999521443642e+38,
                -1.179999945774631e-38,
                float(0),
                1.179999945774631e-38,
                3.3999999521443642e+38,

                None,
                -3.3999999521443642e+38,
                -1.179999945774631e-38,
                float(0),
                1.179999945774631e-38,
                3.3999999521443642e+38,

                None,
                -1.79e+308,
                -2.23e-308,
                float(0),
                2.23e-308,
                1.79e+308,
            )
        )

    def test_numeric(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(NUMERIC(5,3), NULL),
                CONVERT(NUMERIC(5,3), '12.345'),
                CONVERT(NUMERIC(5,3), '12.34567'),
                CONVERT(NUMERIC(5,3), '12.34543'),
                CONVERT(NUMERIC(5,3), 0),
                CONVERT(NUMERIC(5,3), 66),

                CONVERT(DECIMAL(5,3), NULL),
                CONVERT(DECIMAL(5,3), '12.345'),
                CONVERT(DECIMAL(5,3), '12.34567'),
                CONVERT(DECIMAL(5,3), '12.34543'),
                CONVERT(DECIMAL(5,3), 0),
                CONVERT(DECIMAL(5,3), 66)
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                Decimal('12.345'),
                Decimal('12.346'),
                Decimal('12.345'),
                Decimal('0.000'),
                Decimal('66.000'),

                None,
                Decimal('12.345'),
                Decimal('12.346'),
                Decimal('12.345'),
                Decimal('0.000'),
                Decimal('66.000'),
            )
        )

    def test_money(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(MONEY, NULL),
                CONVERT(MONEY, '-922,337,203,685,477.5808'),
                CONVERT(MONEY, '922,337,203,685,477.5807'),

                CONVERT(SMALLMONEY, NULL),
                CONVERT(SMALLMONEY, '-214,748.3648'),
                CONVERT(SMALLMONEY, '214,748.3647')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                self.round_money(Decimal('-922337203685477.5808')),
                self.round_money(Decimal('922337203685477.5807')),

                None,
                # SMALLMONEY seems to be rounded properly by FreeTDS ...
                Decimal('-214748.3648'),
                Decimal('214748.3647'),
            )
        )

    def test_date(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(DATE, NULL),
                CONVERT(DATE, '0001-01-01'),
                CONVERT(DATE, '9999-12-31')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                (
                    '0001-01-01'
                    if self.connection.tds_version < '7.3'
                    else date(1, 1, 1)
                ),
                (
                    '9999-12-31'
                    if self.connection.tds_version < '7.3'
                    else date(9999, 12, 31)
                ),
            )
        )

    def test_time(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(TIME, NULL),
                CONVERT(TIME, '01:02:03.01'),
                CONVERT(TIME, '23:59:59.99')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                (
                    '01:02:03.0100000'
                    if self.connection.tds_version < '7.3'
                    else time(1, 2, 3, 10000)
                ),
                (
                    '23:59:59.9900000'
                    if self.connection.tds_version < '7.3'
                    else time(23, 59, 59, 990000)
                ),
            )
        )

    def test_datetime(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(DATETIME, NULL),
                CONVERT(DATETIME, '1753-1-1'),
                CONVERT(DATETIME, '9999-12-31 23:59:59.997')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                datetime(1753, 1, 1),
                datetime(9999, 12, 31, 23, 59, 59, 997000)
            )
        )

    def test_smalldatetime(self):
        self.cursor.execute(
            # SMALLDATETIME only has minute accuracy.
            '''
            SELECT
                CONVERT(SMALLDATETIME, NULL),
                CONVERT(SMALLDATETIME, '1900-1-1'),
                CONVERT(SMALLDATETIME, '1999-12-31 10:11'),
                CONVERT(SMALLDATETIME, '2076-6-6 23:59:59')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                datetime(1900, 1, 1),
                datetime(1999, 12, 31, 10, 11),
                datetime(2076, 6, 7)
            )
        )

    def test_datetime2(self):
        self.cursor.execute(
            '''
            SELECT
                CONVERT(DATETIME2, NULL),
                CONVERT(DATETIME2, '1-1-1'),
                CONVERT(DATETIME2, '9999-12-31 23:59:59.9999999')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                (
                    '2001-01-01 00:00:00.0000000'
                    if self.connection.tds_version < '7.3'
                    else datetime(2001, 1, 1)
                ),
                (
                    '9999-12-31 23:59:59.9999999'
                    if self.connection.tds_version < '7.3'
                    else (
                        datetime(9999, 12, 31, 23, 59, 59, 999999)
                        if self.tdsdatetime2_supported else
                        datetime(9999, 12, 31, 23, 59, 59, 997 * 1000)
                    )
                ),
            )
        )

    def test_datetimeoffset(self):
        from datetime import timezone, timedelta
        self.cursor.execute(
            '''
            SELECT
                CONVERT(DATETIMEOFFSET, NULL),
                CONVERT(DATETIMEOFFSET, '2023-06-15 12:30:45.1234567 +00:00'),
                CONVERT(DATETIMEOFFSET, '2023-06-15 12:30:45.0000000 +05:30'),
                CONVERT(DATETIMEOFFSET, '2023-06-15 12:30:45.0000000 -05:00')
            '''
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                datetime(2023, 6, 15, 12, 30, 45, 123456,
                         tzinfo=timezone.utc),
                datetime(2023, 6, 15, 12, 30, 45, 0,
                         tzinfo=timezone(timedelta(hours=5, minutes=30))),
                datetime(2023, 6, 15, 12, 30, 45, 0,
                         tzinfo=timezone(timedelta(hours=-5))),
            )
        )

    def test_guid(self):

        uuid1 = uuid.uuid4()
        uuid2 = uuid.uuid4()

        self.cursor.execute(
            '''
            SELECT
                CONVERT(UNIQUEIDENTIFIER, NULL),
                CONVERT(UNIQUEIDENTIFIER, :0),
                CONVERT(UNIQUEIDENTIFIER, :1)
            ''',
            (
                uuid1,
                uuid2
            )
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                uuid1,
                uuid2
            )
        )

    def test_xml(self):

        xml = '<foo><bar>1</bar><baz/></foo>'
        self.cursor.execute(
            '''
            SELECT
                CONVERT(XML, NULL),
                CONVERT(XML, :0)
            ''',
            (
                xml,
            )
        )
        self.assertEqual(
            tuple(self.cursor.fetchone()),
            (
                None,
                xml
            )
        )

    def test_unsupported(self):
        obj = object()
        try:
            self.cursor.execute('SELECT :0', (obj,))
        except ctds.InterfaceError as ex:
            self.assertEqual(
                str(ex),
                'could not implicitly convert Python type "{0}" to SQL'.format(type(obj))
            )
        else:
            self.fail('.execute() did not fail as expected') # pragma: nocover



#!/usr/bin/env python
"""
Minimal setup.py — retained only for the C extension build logic that
cannot be expressed declaratively in pyproject.toml (platform-specific
compiler flags, env-var directory lookups, coverage/profiling toggles).

All package metadata now lives in pyproject.toml.
"""

import glob
import os
import os.path
import platform
import sys

import setuptools

# ---------------------------------------------------------------------------
# Version — single source of truth is pyproject.toml [project] version.
# These constants are compiled into the C extension via define_macros.
# ---------------------------------------------------------------------------
CTDS_MAJOR_VERSION = 2
CTDS_MINOR_VERSION = 1
CTDS_PATCH_VERSION = 0

# ---------------------------------------------------------------------------
# Build flags driven by environment variables
# ---------------------------------------------------------------------------
STRICT = os.environ.get('CTDS_STRICT')
WINDOWS = platform.system() == 'Windows'
COVERAGE = os.environ.get('CTDS_COVER', False)

LIBRARIES = [
    'sybdb',
    'ct',  # required for ct_config only
]

EXTRA_COMPILE_ARGS = []
EXTRA_LINK_ARGS = []

if not WINDOWS:
    if STRICT:
        EXTRA_COMPILE_ARGS += [
            '-ansi',
            '-Wall',
            '-Wextra',
            '-Werror',
            '-Wconversion',
            '-Wpedantic',
            '-std=c99',
        ]

    if os.environ.get('CTDS_PROFILE'):
        EXTRA_COMPILE_ARGS.append('-pg')
        profile_dir = os.environ['CTDS_PROFILE']
        if os.path.isdir(profile_dir):
            EXTRA_COMPILE_ARGS += ['-fprofile-dir', '"{0}"'.format(profile_dir)]
        EXTRA_LINK_ARGS.append('-pg')

    if COVERAGE:
        EXTRA_COMPILE_ARGS += ['-fprofile-arcs', '-ftest-coverage']
        EXTRA_LINK_ARGS.append('-fprofile-arcs')

    # pthread is required on macOS for thread-local storage support.
    if sys.platform == 'darwin':
        EXTRA_LINK_ARGS.append('-lpthread')
else:
    if STRICT:
        EXTRA_COMPILE_ARGS += [
            '/WX',
            '/wd4068',
            '/w14242',
            '/w14254',
            '/w14263',
            '/w14265',
            '/w14287',
            '/we4289',
            '/w14296',
            '/w14311',
            '/w14545',
            '/w14546',
            '/w14547',
            '/w14549',
            '/w14555',
            '/w14619',
            '/w14640',
            '/w14826',
            '/w14905',
            '/w14906',
            '/w14928',
            '/Zi',
        ]
    if COVERAGE:
        EXTRA_COMPILE_ARGS.append('/Od')
        EXTRA_LINK_ARGS.append('/DEBUG')

    LIBRARIES += [
        'shell32',
        'ws2_32',
    ]


def splitdirs(name):
    dirs = os.environ.get(name)
    return dirs.split(os.pathsep) if dirs else []


setuptools.setup(
    ext_modules=[
        setuptools.Extension(
            'k_ctds._tds',
            glob.glob(os.path.join('src', 'k_ctds', '*.c')),
            define_macros=[
                ('CTDS_MAJOR_VERSION', CTDS_MAJOR_VERSION),
                ('CTDS_MINOR_VERSION', CTDS_MINOR_VERSION),
                ('CTDS_PATCH_VERSION', CTDS_PATCH_VERSION),
                ('PY_SSIZE_T_CLEAN', '1'),
                ('MSDBLIB', '1'),
            ],
            include_dirs=splitdirs('CTDS_INCLUDE_DIRS'),
            library_dirs=splitdirs('CTDS_LIBRARY_DIRS'),
            # runtime_library_dirs is not supported on Windows.
            runtime_library_dirs=[] if WINDOWS else splitdirs('CTDS_RUNTIME_LIBRARY_DIRS'),
            extra_compile_args=EXTRA_COMPILE_ARGS,
            extra_link_args=EXTRA_LINK_ARGS,
            libraries=LIBRARIES,
            language='c',
        ),
    ],
)



import datetime
import platform
import time

import unittest

import k_ctds as ctds

class TestTdsDate(unittest.TestCase):

    def test_date(self):
        val = ctds.Date(1, 2, 3)
        self.assertEqual(val, datetime.date(1, 2, 3))

        val = ctds.Date(9999, 12, 31)
        self.assertEqual(val, datetime.date(9999, 12, 31))

    def test_fromticks(self):
        cases = [
            86400,
            123456789,
            time.time()
        ]
        if platform.system() != 'Windows': # pragma: nobranch
            cases += [-1, 0]

        for case in cases:
            self.assertEqual(
                ctds.DateFromTicks(case),
                datetime.datetime.fromtimestamp(case).date()
            )

    def test_fromticks_typeerror(self):
        for case in (None, '', b''):
            self.assertRaises(TypeError, ctds.DateFromTicks, case)

    def test_typeerror(self):
        for args in (
                (),
                (1,),
                (1, 2),
                (None, None, None),
                ('', 1, 2),
                (1, '', 2),
                (1, 1, ''),
        ):
            self.assertRaises(TypeError, ctds.Date, *args)

    def test_valueerror(self):
        for args, part in (
                ((0, 1, 1), 'year'),
                ((1, 0, 1), 'month'),
                ((1, 1, 0), 'day'),
                ((10000, 1, 1), 'year'),
                ((9999, 13, 1), 'month'),
                ((9999, 12, 32), 'day'),
                ((-1, 12, 1), 'year'),
                ((1, -1, 1), 'month'),
                ((1, 1, -1), 'day'),
        ):
            try:
                ctds.Date(*args)
            except ValueError as ex:
                self.assertEqual(str(ex), '{0} is out of range'.format(part))
            else:
                self.fail('.Date() did not fail as expected') # pragma: nocover



import warnings

from .base import TestExternalDatabase

class TestCursorNext(TestExternalDatabase):
    '''Unit tests related to the Cursor.__iter__() member.
    '''

    def test_next(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @test_fetchone TABLE(i INT);
                        INSERT INTO @test_fetchone(i) VALUES (1),(2),(3);
                        SELECT * FROM @test_fetchone;
                        SELECT i * 2 FROM @test_fetchone;
                    '''
                )
                with warnings.catch_warnings(record=True) as warns:
                    self.assertEqual([tuple(row) for row in cursor], [(1,), (2,), (3,)])
                    self.assertEqual(len(warns), 1)
                    self.assertEqual(
                        [str(warn.message) for warn in warns],
                        ['DB-API extension cursor.__iter__() used'] * len(warns)
                    )
                    self.assertEqual(
                        [warn.category for warn in warns],
                        [Warning] * len(warns)
                    )

                self.assertEqual(cursor.nextset(), True)

                with warnings.catch_warnings(record=True) as warns:
                    self.assertEqual([tuple(row) for row in cursor], [(2,), (4,), (6,)])
                    self.assertEqual(len(warns), 1)
                    self.assertEqual(
                        [str(warn.message) for warn in warns],
                        ['DB-API extension cursor.__iter__() used'] * len(warns)
                    )
                    self.assertEqual(
                        [warn.category for warn in warns],
                        [Warning] * len(warns)
                    )

                self.assertEqual(cursor.nextset(), None)

    def test_next_warning_as_error(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @test_fetchone TABLE(i INT);
                        INSERT INTO @test_fetchone(i) VALUES (1),(2),(3);
                        SELECT * FROM @test_fetchone;
                        SELECT i * 2 FROM @test_fetchone;
                    '''
                )
                with warnings.catch_warnings():
                    warnings.simplefilter('error')
                    try:
                        self.assertEqual([tuple(row) for row in cursor], [(1,), (2,), (3,)])
                    except Warning as warn:
                        self.assertEqual('DB-API extension cursor.__iter__() used', str(warn))
                    else:
                        self.fail('.__iter__() did not fail as expected') # pragma: nocover



'''
*cTDS* is a pure C implementation of the `Python DB API 2.0 <https://www.python.org/dev/peps/pep-0249>`_
specification for Microsoft SQL Server.
'''
import logging

# pylint: disable=no-name-in-module,redefined-builtin
from k_ctds._tds import (
    apilevel,
    connect,
    paramstyle,
    threadsafety,

    Date,
    Time,
    Timestamp,
    DateFromTicks,
    TimeFromTicks,
    TimestampFromTicks,
    Binary,

    # DB API 2.0 Exceptions
    Warning,
    Error,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
    InternalError,
    ProgrammingError,
    NotSupportedError,

    Parameter,
    Row,
    RowList,

    # Types
    TDSCHAR as CHAR,
    TDSVARCHAR as VARCHAR,
    TDSNCHAR as NCHAR,
    TDSNVARCHAR as NVARCHAR,
    TDSTEXT as TEXT,
    TDSNTEXT as NTEXT,
    TDSBIT as BIT,
    TDSTINYINT as TINYINT,
    TDSSMALLINT as SMALLINT,
    TDSINT as INT,
    TDSBIGINT as BIGINT,
    TDSFLOAT as FLOAT,
    TDSREAL as REAL,
    TDSDATETIME as DATETIME,
    TDSSMALLDATETIME as SMALLDATETIME,
    TDSDATE as DATE,
    TDSTIME as TIME,
    TDSDATETIME2 as DATETIME2,
    TDSDATETIMEOFFSET as DATETIMEOFFSET,
    TDSIMAGE as IMAGE,
    TDSSMALLMONEY as SMALLMONEY,
    TDSMONEY as MONEY,
    TDSMONEYN as MONEYN,
    TDSNUMERIC as NUMERIC,
    TDSDECIMAL as DECIMAL,
    TDSBINARY as BINARY,
    TDSVARBINARY as VARBINARY,
    TDSGUID as GUID,
    TDSVOID as VOID,

    freetds_version,
    version_info,

    Connection,
    Cursor,

    SqlBigInt,
    SqlBinary,
    SqlChar,
    SqlDate,
    SqlDecimal,
    SqlNVarChar,
    SqlInt,
    SqlSmallInt,
    SqlTinyInt,
    SqlVarBinary,
    SqlVarChar,
)


class NullHandler(logging.Handler):
    def emit(self, record):
        pass # pragma: nocover

    def handle(self, record):
        pass # pragma: nocover

    def createLock(self):
        self.lock = None


# Configure a NullHandler for library logging.
# See https://docs.python.org/2/howto/logging.html#library-config.
logging.getLogger(__name__).addHandler(NullHandler())

del NullHandler


# Define the library version, in major.minor.patch format.
__version__ = '.'.join(str(item) for item in version_info) # pylint: disable=invalid-name



import k_ctds as ctds

from .base import TestExternalDatabase
from .compat import unicode_

class TestConnectionTDSVersion(TestExternalDatabase):
    '''Unit tests related to the Connection.tds_version attribute.
    '''

    def test___doc__(self):
        self.assertEqual(
            ctds.Connection.tds_version.__doc__,
            '''\
The TDS version in use for the connection or :py:data:`None` if the
connection is closed.

:rtype: str
'''
        )

    def test_read(self):
        supports73 = False
        try:
            with self.connect(tds_version='7.3'):
                supports73 = True
        except ctds.InterfaceError: # pragma: nocover
            pass

        versions = [
            ('7.0', '7.0'),
            ('7.1', '7.1'),

            # FreeTDS versions without 7.3 support always downgrade 7.2 to 7.1.
            ('7.2', '7.2' if supports73 else '7.1'),
        ]
        if supports73: # pragma: nobranch
            versions.append(('7.3', '7.3'))

        for tds_version, expected in versions:
            with self.connect(tds_version=tds_version) as connection:
                self.assertTrue(isinstance(connection.tds_version, unicode_))
                self.assertEqual(connection.tds_version, expected)

        self.assertEqual(connection.tds_version, None)

    def test_write(self):
        with self.connect() as connection:
            try:
                connection.tds_version = 9
            except AttributeError:
                pass
            else:
                self.fail('.tds_version did not fail as expected') # pragma: nocover

        try:
            connection.tds_version = None
        except AttributeError:
            pass
        else:
            self.fail('.tds_version did not fail as expected') # pragma: nocover



from datetime import date, datetime, time
from decimal import Decimal
import uuid
import warnings

import k_ctds as ctds

from .base import TestExternalDatabase
from .compat import PY27, PY3, unichr_, unicode_

class TestCursorCallproc(TestExternalDatabase): # pylint: disable=too-many-public-methods
    '''Unit tests related to the Cursor.callproc() method.
    '''

    def test___doc__(self):
        self.assertEqual(
            ctds.Cursor.callproc.__doc__,
            '''\
callproc(sproc, parameters)

Call a stored database procedure with the given name. The sequence of
parameters must contain one entry for each argument that the procedure
expects. The result of the call is returned as modified copy of the input
sequence. Input parameters are left untouched. Output and input/output
parameters are replaced with output values.

.. warning:: Due to `FreeTDS` implementation details, stored procedures
    with both output parameters and resultsets are not supported.

.. warning:: Currently `FreeTDS` does not support passing empty string
    parameters. Empty strings are converted to `NULL` values internally
    before being transmitted to the database.

:pep:`0249#callproc`

:param str sproc: The name of the stored procedure to execute.
:param parameters: Parameters to pass to the stored procedure.
    Parameters passed in a :py:class:`dict` must map from the parameter
    name to value and start with the **@** character. Parameters passed
    in a :py:class:`tuple` are passed in the tuple order.
:type parameters: dict or tuple
:return: The input `parameters` with any output parameters replaced with
    the output values.
:rtype: dict or tuple
'''
        )

    def test_closed_connection(self):
        connection = self.connect()
        with connection.cursor() as cursor:
            connection.close()
            try:
                cursor.callproc('myproc', ())
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'connection closed')
            else:
                self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_invalid_arguments(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                self.assertRaises(TypeError, cursor.callproc)
                self.assertRaises(TypeError, cursor.callproc, 'sproc')
                for case in ('1', None, 1, object()):
                    self.assertRaises(TypeError, cursor.callproc, 'sproc', case)

    def test_unknown_stored_procedure(self):
        with self.connect() as connection:
            sproc = 'this_is_an_unknown_sproc'
            with connection.cursor() as cursor:
                try:
                    cursor.callproc(sproc, ())
                except ctds.ProgrammingError as ex:
                    self.assertEqual(str(ex), 'Could not find stored procedure \'{0}\'.'.format(sproc))
                else:
                    self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_sql_exception(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_exception.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                    AS
                        RAISERROR (N'some custom error %s', 12, 111, 'hello!');
                    '''
                ):
                    try:
                        cursor.callproc(sproc, ())
                    except ctds.ProgrammingError as ex:
                        msg = "some custom error hello!"
                        self.assertEqual(str(ex), msg)
                        self.assertEqual(ex.severity, 12)
                        self.assertEqual(ex.os_error, None)
                        self.assertTrue(self.server_name_and_instance in ex.last_message.pop('server'))
                        self.assertEqual(ex.last_message, {
                            'number': 50000,
                            'state': 111,
                            'severity': 12,
                            'description': msg,
                            'proc': sproc,
                            'line': 4,
                        })
                    else:
                        self.fail('.callproc() did not fail as expected') # pragma: nocover

                # The cursor should be usable after an exception.
                cursor.execute('SELECT @@VERSION')
                self.assertTrue(cursor.fetchall())

    def test_sql_raiseerror_warning(self):
        # The error is simply reported as a warning because another statement follows
        # and completes successfully.
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_raiseerror_warning.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                    AS
                        RAISERROR (N'some custom error %s', 10, -1, 'hello!');
                        SELECT 1 As SomeResult
                    '''
                ):
                    with warnings.catch_warnings(record=True) as warns:
                        cursor.callproc(sproc, ())
                        msg = unicode_('some custom error hello!')
                        self.assertEqual(len(warns), 1)
                        for warn in warns:
                            self.assertIn(msg, str(warn.message))
                        self.assertEqual(
                            [warn.category for warn in warns],
                            [ctds.Warning] * len(warns)
                        )

                    expected = [
                        {
                            'description': msg,
                            'line': 4,
                            'number': 50000,
                            'proc': sproc,
                            'severity': 0, # SQL Server translates 10 -> 0
                            'state': 1,
                        },
                    ]

                    with warnings.catch_warnings(record=True):
                        for index, message in enumerate(connection.messages):
                            self.assertTrue(self.server_name_and_instance in message.pop('server'))
                            self.assertEqual(expected[index], message)

                    self.assertEqual(
                        [tuple(row) for row in cursor.fetchall()],
                        [
                            (1,)
                        ]
                    )

                # The cursor should be usable after a warning.
                with warnings.catch_warnings(record=True) as warns:
                    cursor.execute('SELECT @@VERSION')
                    self.assertTrue(cursor.fetchall())
                    self.assertEqual(len(warns), 0)

    def test_warning_as_error(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_warning_as_error.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                    AS
                        RAISERROR (N'some custom error %s', 10, -1, 'hello!');
                        SELECT 1 As SomeResult
                    '''
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter('error', ctds.Warning)
                        try:
                            cursor.callproc(sproc, ())
                        except ctds.Warning as warn:
                            self.assertIn('some custom error hello!', str(warn))
                        else:
                            self.fail('.callproc() did not fail as expected') # pragma: nocover

                # The cursor should be usable after a warning.
                with warnings.catch_warnings(record=True) as warns:
                    cursor.execute('SELECT @@VERSION')
                    self.assertTrue(cursor.fetchall())
                    self.assertEqual(len(warns), 0)

    def test_sql_print(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_print.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pMsg VARCHAR(MAX)
                    AS
                        PRINT(@pMsg);
                    '''
                ):
                    msg = unicode_('hello world!')
                    with warnings.catch_warnings(record=True) as warns:
                        cursor.callproc(sproc, (msg,))
                        self.assertEqual(len(warns), 0)

                    expected = [
                        {
                            'description': msg,
                            'line': 5,
                            'number': 0,
                            'proc': sproc,
                            'severity': 0,
                            'state': 1,
                        },
                    ]

                    with warnings.catch_warnings(record=True):
                        for index, message in enumerate(connection.messages):
                            self.assertTrue(self.server_name_and_instance in message.pop('server'))
                            self.assertEqual(expected[index], message)

                # The cursor should be usable after a warning.
                with warnings.catch_warnings(record=True) as warns:
                    cursor.execute('SELECT @@VERSION')
                    self.assertTrue(cursor.fetchall())
                    self.assertEqual(len(warns), 0)

    def test_dataerror(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_dataerror.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pInt TINYINT OUTPUT
                    AS
                        SELECT @pInt = @pInt * 255;
                    '''
                ):
                    try:
                        cursor.callproc(sproc, (256,))
                    except ctds.DataError as ex:
                        msg = "Error converting data type smallint to tinyint."
                        self.assertEqual(str(ex), msg)
                        self.assertEqual(ex.severity, 16)
                        self.assertEqual(ex.os_error, None)
                        self.assertTrue(self.server_name_and_instance in ex.last_message.pop('server'))
                        self.assertEqual(ex.last_message, {
                            'number': 8114,
                            'state': 5,
                            'severity': 16,
                            'description': msg,
                            'proc': sproc,
                            'line': 0,
                        })
                    else:
                        self.fail('.callproc() did not fail as expected') # pragma: nocover

                # The cursor should be usable after a warning.
                with warnings.catch_warnings(record=True) as warns:
                    cursor.execute('SELECT @@VERSION')
                    self.assertTrue(cursor.fetchall())
                    self.assertEqual(len(warns), 0)

    def test_sql_conversion_warning(self):
        with self.connect(ansi_defaults=False) as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_conversion_warning.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pInt TINYINT OUTPUT,
                        @pBigInt BIGINT OUTPUT
                    AS
                        SELECT @pInt = @pInt * 255;
                        SELECT @pBigInt = @pBigInt - 1000;
                    '''
                ):
                    with warnings.catch_warnings(record=True) as warns:
                        inputs = (
                            ctds.Parameter(10, output=True),
                            ctds.Parameter(1234, output=True),
                        )
                        outputs = cursor.callproc(sproc, inputs)
                        self.assertEqual(outputs, (None, inputs[1].value - 1000))
                        self.assertEqual(len(warns), 1)
                        msg = 'Arithmetic overflow occurred.'
                        for warn in warns:
                            self.assertIn(msg, str(warn.message))
                        self.assertEqual(
                            [warn.category for warn in warns],
                            [ctds.Warning] * len(warns)
                        )

    def test_sql_conversion_warning_as_error(self): # pylint: disable=invalid-name
        with self.connect(ansi_defaults=False) as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_conversion_warning_as_error.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pInt TINYINT OUTPUT,
                        @pBigInt BIGINT OUTPUT
                    AS
                        SELECT @pInt = @pInt * 255;
                        SELECT @pBigInt = @pBigInt - 1000;
                    '''
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter('error', ctds.Warning)
                        inputs = (
                            ctds.Parameter(10, output=True),
                            ctds.Parameter(1234, output=True),
                        )
                        try:
                            cursor.callproc(sproc, inputs)
                        except ctds.Warning as warn:
                            self.assertIn('Arithmetic overflow occurred.', str(warn))
                        else:
                            self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_sql_conversion_error(self):
        with self.connect(ansi_defaults=True) as connection:
            with connection.cursor() as cursor:
                sproc = self.test_sql_conversion_warning.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pInt TINYINT OUTPUT,
                        @pBigInt BIGINT OUTPUT
                    AS
                        SELECT @pInt = @pInt * 255;
                        SELECT @pBigInt = @pBigInt - 1000;
                    '''
                ):
                    inputs = (
                        ctds.Parameter(10, output=True),
                        ctds.Parameter(1234, output=True),
                    )
                    try:
                        cursor.callproc(sproc, inputs)
                    except ctds.DataError as ex:
                        self.assertEqual(
                            str(ex),
                            'Arithmetic overflow error for data type tinyint, value = {0}.'.format(
                                inputs[0].value * 255
                            )
                        )

    def test_output_with_resultset(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_output_with_resultset.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pOut INT OUTPUT
                    AS
                        SET @pOut = 5;
                        SELECT 10;
                    '''
                ):
                    with warnings.catch_warnings(record=True) as warns:
                        cursor.callproc(sproc, (ctds.Parameter(ctds.SqlInt(0), output=True),))
                        cursor.callproc(sproc, {'@pOut': ctds.Parameter(ctds.SqlInt(0), output=True)})
                        self.assertEqual(len(warns), 2)
                        self.assertEqual(
                            [str(warn.message) for warn in warns],
                            ['output parameters are not supported with result sets'] * len(warns)
                        )
                        self.assertEqual(
                            [warn.category for warn in warns],
                            [Warning] * len(warns)
                        )

    def test_output_with_resultset_as_error(self): # pylint: disable=invalid-name
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_output_with_resultset_as_error.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pOut INT OUTPUT
                    AS
                        SET @pOut = 5;
                        SELECT 10;
                    '''
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter('error')
                        try:
                            cursor.callproc(sproc, (ctds.Parameter(ctds.SqlInt(0), output=True),))
                        except Warning as warn:
                            self.assertEqual('output parameters are not supported with result sets', str(warn))
                        else:
                            self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_dict_invalid_name(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_dict_invalid_name.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pParameter BIGINT
                    AS
                        SELECT @pParameter;
                    '''
                ):
                    for case in ('pParameter', 1, '1', None, ''):
                        try:
                            cursor.callproc(sproc, {case: 1234})
                        except ctds.InterfaceError as ex:
                            self.assertEqual(
                                str(ex),
                                'invalid parameter name "{0}"'.format(repr(case))
                            )
                        else:
                            self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_dict(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_dict.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pBigInt BIGINT,
                        @pVarChar VARCHAR(32),
                        @pVarBinary VARBINARY(32),
                        @pDateTime DATETIME,
                        @pDecimal DECIMAL(32, 20)
                    AS
                        SELECT
                            @pDecimal AS Decimal,
                            @pVarBinary AS VarBinary,
                            @pVarChar AS VarChar,
                            @pBigInt AS BigInt,
                            @pDateTime AS DateTime;
                    '''
                ):
                    types = [str]
                    if not PY3: # pragma: nocover
                        types.append(unicode_)
                    for type_ in types:
                        inputs = {
                            type_('@pDecimal'): Decimal('5.31234123'),
                            type_('@pBigInt'): 3453456908,
                            type_('@pVarBinary'): b'1234567890',
                            type_('@pVarChar'): 'this is some string',
                            type_('@pDateTime'): datetime(2011, 11, 5, 12, 12, 12)
                        }
                        outputs = cursor.callproc(sproc, inputs)
                        self.assertEqual(inputs, outputs)
                        for key in inputs:
                            # (N)(VAR)CHAR parameters are translated to UCS-2 and therefore
                            # the id() values will never match.
                            if key != '@pVarChar':
                                self.assertEqual(id(outputs[key]), id(inputs[key]))
                            self.assertEqual(outputs[key], inputs[key])

                        self.assertEqual(
                            [tuple(row) for row in cursor.fetchall()],
                            [
                                (
                                    inputs[type_('@pDecimal')],
                                    inputs[type_('@pVarBinary')],
                                    inputs[type_('@pVarChar')],
                                    inputs[type_('@pBigInt')],
                                    inputs[type_('@pDateTime')]
                                )
                            ]
                        )

    def test_dict_outputs(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_dict.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pBigInt BIGINT OUTPUT,
                        @pVarChar VARCHAR(32) OUTPUT,
                        @pVarBinary VARBINARY(32) OUTPUT,
                        @pFloat FLOAT OUTPUT,
                        @pDateTime DATETIME,
                        @pDateTimeOut DATETIME OUTPUT,
                        @pDateTime2 DATETIME2,
                        @pDateTime2Out DATETIME2 OUTPUT,
                        @pDate DATE,
                        @pDateOut DATE OUTPUT,
                        @pTime TIME,
                        @pTimeOut TIME OUTPUT
                    AS
                        SET @pBigInt = @pBigInt * 2;
                        SET @pVarBinary = CONVERT(VARBINARY(32), @pVarChar);
                        SET @pVarChar = CONVERT(VARCHAR(32), @pDateTime, 120);
                        SET @pFloat = @pFloat * 2.1;
                        SET @pDateTimeOut = CONVERT(DATETIME, '2016-02-03 04:05:06.783');
                        SET @pDateTime2Out = CONVERT(DATETIME2, '2017-01-01 01:02:03.1234567');
                        SET @pDateOut = CONVERT(DATE, '2017-01-01');
                        SET @pTimeOut = CONVERT(TIME, '01:02:03.1234567');
                    '''
                ):
                    types = [str]
                    if not PY3: # pragma: nocover
                        types.append(unicode_)
                    for type_ in types:
                        inputs = {
                            type_('@pBigInt'): ctds.Parameter(12345, output=True),
                            type_('@pVarChar'): ctds.Parameter(
                                unicode_('hello world, how\'s it going! '),
                                output=True
                            ),
                            type_('@pVarBinary'): ctds.Parameter(ctds.SqlVarBinary(None, size=32), output=True),
                            type_('@pFloat'): ctds.Parameter(1.23, output=True),
                            type_('@pDateTime'): datetime(2011, 11, 5, 12, 12, 12),
                            type_('@pDateTimeOut'): ctds.Parameter(datetime(2011, 11, 5, 12, 12, 12), output=True),
                            type_('@pDateTime2'): datetime(2011, 11, 5, 12, 12, 12, 999999),
                            type_('@pDateTime2Out'): ctds.Parameter(datetime.utcnow(), output=True),
                            type_('@pDate'): date(2011, 11, 5),
                            type_('@pDateOut'): ctds.Parameter(date.today(), output=True),
                            type_('@pTime'): time(12, 12, 12, 123456),
                            type_('@pTimeOut'): ctds.Parameter(datetime.utcnow().time(), output=True),
                        }
                        outputs = cursor.callproc(sproc, inputs)
                        self.assertNotEqual(id(outputs[type_('@pBigInt')]), id(inputs[type_('@pBigInt')]))
                        self.assertNotEqual(id(outputs[type_('@pVarChar')]), id(inputs[type_('@pVarChar')]))
                        self.assertNotEqual(id(outputs[type_('@pVarBinary')]), id(inputs[type_('@pVarBinary')]))
                        self.assertEqual(id(outputs[type_('@pDateTime')]), id(inputs[type_('@pDateTime')]))
                        self.assertEqual(id(outputs[type_('@pDateTime2')]), id(inputs[type_('@pDateTime2')]))
                        self.assertEqual(id(outputs[type_('@pDate')]), id(inputs[type_('@pDate')]))
                        self.assertEqual(id(outputs[type_('@pTime')]), id(inputs[type_('@pTime')]))

                        self.assertEqual(
                            outputs[type_('@pBigInt')],
                            inputs[type_('@pBigInt')].value * 2
                        )
                        self.assertEqual(
                            outputs[type_('@pVarChar')],
                            inputs[type_('@pDateTime')].strftime('%Y-%m-%d %H:%M:%S')
                        )
                        self.assertEqual(
                            outputs[type_('@pVarBinary')],
                            bytes(inputs[type_('@pVarChar')].value.encode('utf-8'))
                        )
                        self.assertEqual(
                            outputs[type_('@pFloat')],
                            inputs[type_('@pFloat')].value * 2.1
                        )
                        self.assertEqual(
                            outputs[type_('@pDateTimeOut')],
                            datetime(2016, 2, 3, 4, 5, 6, 783000)
                        )
                        self.assertEqual(
                            outputs[type_('@pDateTime2Out')],
                            datetime(
                                2017, 1, 1, 1, 2, 3,
                                123456 if self.tdsdatetime2_supported else 123000
                            )
                        )
                        self.assertEqual(
                            outputs[type_('@pDateOut')],

                            # $future: When passing TDSDATE to FreeTDS is
                            # supported, this should be a `datetime.date`
                            datetime(2017, 1, 1)
                        )
                        self.assertEqual(
                            outputs[type_('@pTimeOut')],
                            time(1, 2, 3, 123456)
                            if self.tdstime_supported else
                            datetime(1900, 1, 1, 1, 2, 3, 123000)
                        )

    def test_binary(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_binary.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pBinary BINARY(8),
                        @pVarBinary VARBINARY(16),

                        @pBinaryNone BINARY(1),
                        @pVarBinaryNone VARBINARY(1)
                    AS
                        SELECT
                            CONVERT(VARCHAR(32), @pBinary, 1),
                            CONVERT(VARCHAR(32), @pVarBinary, 1),

                            @pBinaryNone,
                            @pVarBinaryNone
                    '''
                ):
                    inputs = (
                        b'\x01\x02\x03\x04',
                        bytearray(b'\xde\xad\xbe\xef\xde\xad\xbe\xef'),
                        ctds.SqlVarBinary(None),
                        ctds.SqlVarBinary(None)
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertEqual(id(outputs[1]), id(inputs[1]))

                    self.assertEqual(tuple(cursor.fetchone()), (
                        ('0x0102030400000000', '0xDEADBEEFDEADBEEF', None, None)
                    ))

                    input_dict = {
                        '@pBinaryNone': ctds.SqlVarBinary(None),
                        '@pVarBinary': bytearray(b'\xde\xad\xbe\xef\xde\xad\xbe\xef'),
                        '@pBinary': b'\x01\x02\x03\x04',
                        '@pVarBinaryNone': ctds.SqlVarBinary(None)
                    }
                    outputs = cursor.callproc(sproc, input_dict)
                    self.assertEqual(id(outputs['@pBinary']), id(input_dict['@pBinary']))
                    self.assertEqual(id(outputs['@pVarBinary']), id(input_dict['@pVarBinary']))

                    self.assertEqual(tuple(cursor.fetchone()), (
                        ('0x0102030400000000', '0xDEADBEEFDEADBEEF', None, None)
                    ))

    def test_binary_output(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_binary_output.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pIn VARBINARY(8),
                        @pOut VARBINARY(16) OUTPUT
                    AS
                        SET @pOut = @pIn + @pIn;
                    '''
                ):
                    inputs = (
                        b'\x01\x02\x03\x04',
                        ctds.Parameter(ctds.SqlVarBinary(None, size=16), output=True)
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertEqual(outputs[1], inputs[0] + inputs[0])
                    self.assertRaises(ctds.InterfaceError, cursor.fetchone)

                    # Note: When the output parameter size is less than the actual data,
                    # SQL Server truncates the data.
                    inputs = (
                        b'\x01\x02\x03\x04',
                        ctds.Parameter(
                            ctds.SqlVarBinary(None, size=2),
                            output=True
                        )
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertEqual(outputs[1], inputs[0][:2])
                    self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_binary_inputoutput(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_binary_inputoutput.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pInOut VARBINARY(16) OUTPUT
                    AS
                        SET @pInOut = @pInOut + @pInOut;
                    '''
                ):
                    value = b'\x01\x02\x03\x04'
                    inputs = (
                        ctds.Parameter(
                            ctds.SqlVarBinary(value, size=16),
                            output=True
                        ),
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(outputs, (value + value,))
                    self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_int(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_int.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pTinyInt TINYINT,
                        @pSmallInt SMALLINT,
                        @pInt INT,
                        @pBigInt BIGINT,

                        @pTinyIntOut TINYINT OUTPUT,
                        @pSmallIntOut SMALLINT OUTPUT,
                        @pIntOut INT OUTPUT,
                        @pBigIntOut BIGINT OUTPUT
                    AS
                        SELECT @pTinyIntOut = @pTinyInt + 1;
                        SELECT @pSmallIntOut = -1 * @pSmallInt;
                        SELECT @pIntOut = -1 * @pInt;
                        SELECT @pBigIntOut = -1 * @pBigInt;
                    '''
                ):
                    inputs = (
                        1,
                        2 ** 8,
                        2 ** 15,
                        2 ** 31,
                        ctds.Parameter(ctds.SqlTinyInt(None), output=True),
                        ctds.Parameter(ctds.SqlSmallInt(None), output=True),
                        ctds.Parameter(ctds.SqlInt(None), output=True),
                        ctds.Parameter(ctds.SqlBigInt(None), output=True),
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertRaises(ctds.InterfaceError, cursor.fetchone)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertEqual(id(outputs[1]), id(inputs[1]))
                    self.assertEqual(id(outputs[2]), id(inputs[2]))
                    self.assertEqual(id(outputs[3]), id(inputs[3]))

                    self.assertEqual(
                        outputs[4:],
                        (
                            outputs[0] + 1,
                            outputs[1] * -1,
                            outputs[2] * -1,
                            outputs[3] * -1,
                        )
                    )

    def test_float(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_float.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pReal REAL,
                        @pFloat24 FLOAT(24),
                        @pFloat53 FLOAT(53),

                        @pRealOut REAL OUTPUT,
                        @pFloat24Out FLOAT(24) OUTPUT,
                        @pFloat53Out FLOAT(53) OUTPUT
                    AS
                        SET @pRealOut = @pReal * 2;
                        SET @pFloat24Out = @pFloat24 * 2;
                        SET @pFloat53Out = @pFloat53 * 2;

                        SELECT
                            @pReal AS Real,
                            @pFloat24 AS Float24,
                            @pFloat53 AS Float53;
                    '''
                ):
                    inputs = (
                        12345.6787109375,
                        1234567936.0,
                        453423423423445352345.23424123,
                        0, 0, 0
                    )
                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertEqual(id(outputs[1]), id(inputs[1]))
                    self.assertEqual(id(outputs[2]), id(inputs[2]))

                    self.assertEqual(tuple(cursor.fetchone()), inputs[0:3])
                    self.assertEqual(cursor.fetchone(), None)

    def test_datetime(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_datetime.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pDatetime DATETIME,
                        @pDatetimeOut DATETIME OUTPUT
                    AS
                        SET @pDatetimeOut = '{0}';

                        SELECT
                            @pDatetime AS Datetime;
                    '''.format(datetime(1999, 12, 31, 23, 59, 59).strftime('%Y-%m-%d %H:%M:%S'))
                ):
                    inputs = (
                        datetime(2001, 1, 1, 1, 1, 1, 123000),
                        datetime(1984, 1, 1, 1, 1, 1),
                    )

                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))

                    self.assertEqual(tuple(cursor.fetchone()), inputs[0:1])
                    self.assertEqual(cursor.fetchone(), None)

    def test_smalldatetime(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_datetime.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pSmallDatetime SMALLDATETIME,
                        @pSmallDatetimeOut SMALLDATETIME OUTPUT
                    AS
                        SET @pSmallDatetimeOut = '{0}';
                    '''.format(datetime(1999, 12, 31, 23, 59, 59).strftime('%Y-%m-%d %H:%M:%S'))
                ):
                    inputs = (
                        datetime(2001, 1, 1, 1, 1, 1, 123456),
                        datetime(1984, 1, 1, 1, 1, 1),
                    )

                    outputs = cursor.callproc(sproc, inputs)
                    self.assertEqual(id(outputs[0]), id(inputs[0]))
                    self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_decimal(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_decimal.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pDecimal DECIMAL(7,3),
                        @pDecimalOut VARCHAR(MAX) OUTPUT
                    AS
                        SET @pDecimalOut = CONVERT(VARCHAR(MAX), @pDecimal);
                    '''
                ):
                    for value, string, truncation in (
                            (1000.0, '1000.000', False),
                            (-1000.0, '-1000.000', False),
                            (9999.999, '9999.999', PY27),
                            (-9999.999, '-9999.999', PY27),
                            (.999, '0.999', PY27),
                            (-.999, '-0.999', PY27),
                            (.001, '0.001', PY27),
                            (-.001, '-0.001', PY27),
                            (0, '0.000', False),
                    ):
                        inputs = (
                            Decimal(value if PY27 else str(value)),
                            ctds.Parameter(unicode_('*' * 256), output=True),
                        )

                        with warnings.catch_warnings(record=True) as warns:
                            outputs = cursor.callproc(sproc, inputs)
                            if truncation:
                                self.assertEqual(len(warns), 1)
                                msg = "Decimal('{0}') exceeds SQL DECIMAL precision; truncating".format(inputs[0])
                                self.assertEqual(
                                    [str(warn.message) for warn in warns],
                                    [msg] * len(warns)
                                )
                                self.assertEqual(warns[0].category, Warning)
                            else:
                                self.assertEqual(len(warns), 0)

                            self.assertEqual(id(outputs[0]), id(inputs[0]))
                            self.assertEqual(string, outputs[1])
                            self.assertRaises(ctds.InterfaceError, cursor.fetchone)

    def test_decimal_output(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_decimal_output.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pString VARCHAR(MAX),
                        @pDecimalOut DECIMAL(15,5) OUTPUT
                    AS
                        SET @pDecimalOut = CONVERT(DECIMAL(15,5), @pString);
                    '''
                ):
                    for value in (
                            '1230.456',
                            '123456789.1234',
                            '123',
                            '1234567890.123456789123456789',
                            None
                    ):
                        inputs = (
                            value,
                            ctds.Parameter(ctds.SqlDecimal(value, precision=15, scale=5), output=True),
                        )

                        outputs = cursor.callproc(sproc, inputs)
                        self.assertEqual(id(outputs[0]), id(inputs[0]))
                        if outputs[0] is not None:
                            self.assertEqual(
                                Decimal(outputs[0]).quantize(Decimal('.00001')),
                                outputs[1]
                            )
                        else:
                            self.assertEqual(None, outputs[1])

    def test_decimal_outofrange(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_decimal_outofrange.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pDecimal DECIMAL(7,3),
                        @pDecimalOut VARCHAR(MAX) OUTPUT
                    AS
                        SET @pDecimalOut = CONVERT(VARCHAR(MAX), @pDecimal);
                    '''
                ):
                    for value in (
                            10 ** 38,
                            -10 ** 38,
                    ):
                        inputs = (
                            Decimal(value),
                            ctds.Parameter(unicode_('*' * 256), output=True),
                        )
                        try:
                            cursor.callproc(sproc, inputs)
                        except ctds.DataError as ex:
                            self.assertEqual(
                                str(ex),
                                "Decimal('{0}') out of range".format(value)
                            )
                        else:
                            self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_decimal_internalerror(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_decimal_internalerror.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pDecimal DECIMAL(7,3)
                    AS
                        SELECT CONVERT(VARCHAR(MAX), @pDecimal);
                    '''
                ):
                    inputs = (Decimal('NaN'),)
                    try:
                        cursor.callproc(sproc, inputs)
                    except ctds.InterfaceError as ex:
                        self.assertEqual(
                            str(ex),
                            'failed to convert {0}'.format(repr(inputs[0]))
                        )
                    else:
                        self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_varchar(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_varchar.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar VARCHAR(256),
                        @pVarCharOut VARCHAR(256) OUTPUT
                    AS
                        SET @pVarCharOut = @pVarChar;
                    '''
                ):
                    format_ = (
                        unichr_(191) + unicode_(' 8 ') +
                        unichr_(247) + unicode_(' 2 = 4 ? {0} {1}')
                    )
                    snowman = unichr_(9731)

                    # Python must be built with UCS4 support to test the large codepoints.
                    if self.UCS4_SUPPORTED:
                        catface = unichr_(128568)
                    else:
                        catface = self.UNICODE_REPLACEMENT

                    # Older versions of SQL server don't support passing codepoints outside
                    # of the server's code page. SQL Server defaults to latin-1, so assume
                    # non-latin-1 codepoints won't be supported.
                    if not self.use_sp_executesql: # pragma: nocover
                        catface = unicode_('?')
                        snowman = unicode_('?')

                    inputs = (
                        format_.format(snowman, catface),
                        ctds.Parameter(ctds.SqlVarChar(None, size=256), output=True),
                    )

                    # If the connection supports UTF-16, unicode codepoints outside of the UCS-2
                    # range are supported and not replaced by ctds.
                    if self.use_utf16:
                        outputs = cursor.callproc(sproc, inputs)
                        # SQL server will improperly convert the values to single-byte, which corrupts them.
                        # FreeTDS will replace them on reads with a '?'.
                        # The catface is a multi-byte UTF-16 character, and therefore is replaced by two '?'
                        # in UCS-4 builds, and one '?' in UCS-2 builds of python.
                        self.assertEqual(
                            format_.format(unicode_('?'), unicode_('??' if self.UCS4_SUPPORTED else '?')),
                            outputs[1]
                        )
                    else: # pragma: nocover
                        # The catface is not representable in UCS-2, and therefore is replaced.
                        with warnings.catch_warnings(record=True) as warns:
                            outputs = cursor.callproc(sproc, inputs)
                            if ord(catface) > 2**16:
                                self.assertEqual(len(warns), 1)
                                msg = 'Unicode codepoint U+{0:08X} is not representable in UCS-2; replaced with U+{1:04X}'.format( # pylint: disable=line-too-long
                                    ord(catface),
                                    ord(self.UNICODE_REPLACEMENT)
                                )
                                self.assertEqual(
                                    [str(warn.message) for warn in warns],
                                    [msg] * len(warns)
                                )
                                self.assertEqual(
                                    [warn.category for warn in warns],
                                    [UnicodeWarning] * len(warns)
                                )
                            else:
                                self.assertEqual(len(warns), 0) # pragma: nocover

                        # SQL server will improperly convert the values to single-byte, which corrupts them.
                        # FreeTDS will replace them on reads with a '?'.
                        self.assertEqual(
                            format_.format(unicode_('?'), unicode_('?')),
                            outputs[1]
                        )

                    self.assertEqual(id(inputs[0]), id(outputs[0]))
                    self.assertNotEqual(id(inputs[1]), id(outputs[1]))

    def test_ucs2_warning_as_error(self):
        if not self.use_utf16 and self.use_sp_executesql and self.UCS4_SUPPORTED: # pragma: nocover
            with self.connect() as connection:
                with connection.cursor() as cursor:
                    sproc = self.test_ucs2_warning_as_error.__name__
                    with self.stored_procedure(
                        cursor,
                        sproc,
                        '''
                            @pVarChar VARCHAR(256),
                            @pVarCharOut VARCHAR(256) OUTPUT
                        AS
                            SET @pVarCharOut = @pVarChar;
                        '''
                    ):
                        format_ = (
                            unichr_(191) + unicode_(' 8 ') +
                            unichr_(247) + unicode_(' 2 = 4 ? {0}')
                        )
                        catface = unichr_(128568)

                        inputs = (
                            format_.format(catface),
                            ctds.Parameter(ctds.SqlVarChar(None, size=256), output=True),
                        )

                        with warnings.catch_warnings():
                            warnings.simplefilter('error', UnicodeWarning)
                            try:
                                cursor.callproc(sproc, inputs)
                            except UnicodeWarning as warn:
                                msg = 'Unicode codepoint U+{0:08X} is not representable in UCS-2; replaced with U+{1:04X}'.format( # pylint: disable=line-too-long
                                    ord(catface),
                                    ord(self.UNICODE_REPLACEMENT)
                                )
                                self.assertEqual(msg, str(warn))
                            else:
                                self.fail('.callproc() did not fail as expected') # pragma: nocover

    def test_varchar_bytes(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_varchar_bytes.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar VARCHAR(MAX)
                    AS
                        SELECT LEN(@pVarChar), @pVarChar;
                    '''
                ):
                    inputs = (b'some bytes for a VARCHAR parameter',)
                    cursor.callproc(sproc, inputs)
                    row = cursor.fetchone()
                    self.assertEqual(len(inputs[0]), row[0])
                    self.assertEqual(unicode_(inputs[0], encoding='ascii'), row[1])

    def test_varchar_max(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_varchar_max.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar VARCHAR(MAX)
                    AS
                        SELECT LEN(@pVarChar), @pVarChar;
                    '''
                ):
                    lengths = [
                        1, 2, 3, 3999, 4000
                    ]

                    lengths.extend([
                        4001, 7999, 8000, 8001
                    ])
                    for length in lengths:
                        inputs = (unicode_('*') * length,)
                        cursor.callproc(sproc, inputs)
                        row = cursor.fetchone()
                        self.assertEqual(len(inputs[0]), row[0])
                        self.assertEqual(inputs[0], row[1])

    def test_varchar_null(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_varchar_null.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar VARCHAR(256),
                        @pIsNullOut BIT OUTPUT
                    AS
                        IF @pVarChar IS NULL
                            BEGIN
                                SET @pIsNullOut = 1;
                            END
                        ELSE
                            BEGIN
                                SET @pIsNullOut = 0;
                            END
                    '''
                ):
                    for value in (None, unicode_(''), unicode_('0'), unicode_('one')):
                        inputs = (
                            value,
                            ctds.Parameter(True, output=True),
                        )
                        outputs = cursor.callproc(sproc, inputs)

                        # $future: fix this once supported by FreeTDS
                        # Currently FreeTDS (really the db-lib API) will
                        # turn the empty string to NULL
                        self.assertEqual(outputs[1], inputs[0] is None or inputs[0] == '')

    def test_varchar_empty(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_varchar_null.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar VARCHAR(256)
                    AS
                        SELECT @pVarChar, LEN(@pVarChar) AS Length
                    '''
                ):
                    for value in (None, unicode_(''), unicode_('0'), unicode_('one')):
                        inputs = (
                            value,
                        )
                        cursor.callproc(sproc, inputs)
                        self.assertEqual(
                            tuple(cursor.fetchone()),
                            (None, None) if value in (None, unicode_('')) else (value, len(value))
                        )

    def test_nvarchar(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_nvarchar.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar NVARCHAR(256),
                        @pVarCharOut NVARCHAR(256) OUTPUT
                    AS
                        SET @pVarCharOut = @pVarChar;
                    '''
                ):
                    format_ = (
                        unichr_(191) + unicode_(' 8 ') +
                        unichr_(247) + unicode_(' 2 = 4 ? {0} {1} {2} test')
                    )
                    snowman = unichr_(9731)

                    # Python must be built with UCS4 support to test the large codepoints.
                    if self.UCS4_SUPPORTED:
                        catface = unichr_(128568)
                        flower = unichr_(127802)
                    else:
                        catface = self.UNICODE_REPLACEMENT
                        flower = self.UNICODE_REPLACEMENT

                    # Older versions of SQL server don't support passing codepoints outside
                    # of the server's code page. SQL Server defaults to latin-1, so assume
                    # non-latin-1 codepoints won't be supported.
                    if not self.use_sp_executesql: # pragma: nocover
                        catface = unicode_('?')
                        snowman = unicode_('?')
                        flower = unicode_('?')

                    inputs = (
                        format_.format(snowman, catface, flower),
                        ctds.Parameter(ctds.SqlVarChar(None, size=256), output=True),
                    )

                    # If the connection supports UTF-16, unicode codepoints outside of the UCS-2
                    # range are supported and not replaced by ctds.
                    if self.use_utf16:
                        outputs = cursor.callproc(sproc, inputs)
                        self.assertEqual(inputs[0], outputs[1])
                    else: # pragma: nocover
                        # The catface is not representable in UCS-2, and therefore is replaced.
                        with warnings.catch_warnings(record=True) as warns:
                            outputs = cursor.callproc(sproc, inputs)
                            if ord(catface) > 2**16:
                                self.assertEqual(len(warns), 2)
                                msg = unicode_('Unicode codepoint U+{0:08X} is not representable in UCS-2; replaced with U+{1:04X}') # pylint: disable=line-too-long
                                self.assertEqual(
                                    [str(warn.message) for warn in warns],
                                    [msg.format(ord(char), ord(self.UNICODE_REPLACEMENT)) for char in (catface, flower)]
                                )
                                self.assertEqual(
                                    [warn.category for warn in warns],
                                    [UnicodeWarning] * len(warns)
                                )
                            else:
                                self.assertEqual(len(warns), 0) # pragma: nocover

                        self.assertEqual(
                            format_.format(
                                snowman,
                                self.UNICODE_REPLACEMENT if self.use_sp_executesql else unicode_('?'),
                                self.UNICODE_REPLACEMENT if self.use_sp_executesql else unicode_('?')
                            ),
                            outputs[1]
                        )

                    self.assertEqual(id(inputs[0]), id(outputs[0]))
                    self.assertNotEqual(id(inputs[1]), id(outputs[1]))

    def test_nvarchar_max(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_nvarchar_max.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pVarChar NVARCHAR(MAX)
                    AS
                        SELECT LEN(@pVarChar), @pVarChar;
                    '''
                ):
                    lengths = [
                        1, 2, 3, 3999, 4000
                    ]
                    # FreeTDS 0.92.x doesn't properly handle NVARCHAR types > 4000 characters.
                    if self.freetds_version[:2] != (0, 92): # pragma: nocover
                        lengths.extend([
                            4001, 7999, 8000, 8001
                        ])

                    for length in lengths:
                        inputs = (
                            unicode_(b'\xe3\x83\x9b' if self.nchars_supported else b'&', encoding='utf-8') * length,
                        )
                        cursor.callproc(sproc, inputs)
                        row = cursor.fetchone()
                        self.assertEqual(len(inputs[0]), row[0])
                        self.assertEqual(inputs[0], row[1])

    def test_nvarchar_null(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_nvarchar_null.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pNVarChar NVARCHAR(256),
                        @pIsNullOut BIT OUTPUT
                    AS
                        IF @pNVarChar IS NULL
                            BEGIN
                                SET @pIsNullOut = 1;
                            END
                        ELSE
                            BEGIN
                                SET @pIsNullOut = 0;
                            END
                    '''
                ):
                    for value in (None, unicode_(''), unicode_('0'), unicode_('one')):
                        inputs = (
                            value,
                            ctds.Parameter(True, output=True),
                        )
                        outputs = cursor.callproc(sproc, inputs)
                        self.assertEqual(outputs[1], inputs[0] in (None, unicode_('')))

    def test_guid(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_guid.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                        @pIn UNIQUEIDENTIFIER,
                        @pOut UNIQUEIDENTIFIER OUTPUT
                    AS
                        SET @pOut = @pIn;
                    '''
                ):
                    for value in (
                            None,
                            uuid.uuid1(),
                    ):
                        inputs = (
                            value,
                            ctds.Parameter(uuid.uuid1(), output=True)
                        )
                        outputs = cursor.callproc(sproc, inputs)

                        # FreeTDS doesn't support passing raw GUIDs, so they are convereted to
                        # VARCHAR. On output, the type is also VARCHAR.
                        expected = uuid.UUID('{{{0}}}'.format(outputs[1])) if value is not None else None
                        self.assertEqual(value, expected)



import k_ctds as ctds

from .base import TestExternalDatabase

class TestCursorNextset(TestExternalDatabase):
    '''Unit tests related to the Cursor.nextset() method.
    '''
    def test___doc__(self):
        doc = '''\
nextset()

Skip to the next available set, discarding any remaining rows from the
current set.

:pep:`0249#nextset`

:return:
    :py:data:`True` if there was another result set or :py:data:`None`
    if not.
:rtype: bool
'''
        self.assertEqual(
            ctds.Cursor.nextset.__doc__,
            doc
        )

    def test_closed(self):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.close()
            try:
                cursor.nextset()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'cursor closed')
            else:
                self.fail('.nextset() did not fail as expected') # pragma: nocover

    def test_closed_connection(self):
        connection = self.connect()
        with connection.cursor() as cursor:
            connection.close()
            try:
                cursor.nextset()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'connection closed')
            else:
                self.fail('.nextset() did not fail as expected') # pragma: nocover

    def test_partial_read(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                    DECLARE @tTable TABLE (i int);
                    INSERT INTO
                        @tTable
                    VALUES
                        (1),
                        (2),
                        (3),
                        (4);
                    SELECT * FROM @tTable;
                    SELECT 'resultset 2';
                    '''
                )

                self.assertTrue(cursor.nextset())
                self.assertEqual(
                    [tuple(row) for row in cursor.fetchall()],
                    [('resultset 2',)]
                )



from contextlib import contextmanager
from collections import namedtuple
import threading
import time
import warnings

PooledConnection = namedtuple('PooledConnection', ['connection', 'released'])


class ConnectionPool(object): # pylint: disable=too-many-instance-attributes,useless-object-inheritance
    '''
    A basic connection pool for DB-API 2.0-compliant database drivers.

    Example usage:

    .. code-block:: python

        config = {
            'server': 'my-host',
            'database': 'MyDefaultDatabase',
            'user': 'my-username',
            'password': 'my-password',
            'appname': 'my-app-name',
            'timeout': 5
        }
        pool = ConnectionPool(pydbmod, config)

        conn = pool.acquire()
        try:
            cursor = conn.cursor()
            try:
                cursor.execute('SELECT @@VERSION')
            finally:
                cursor.close()
        finally:
            pool.release(conn)


    .. versionadded:: 1.2

    .. note:: Idle connection collection only occurs during calls to
        :py:meth:`.acquire()` and does not happen automatically in the
        background.

    :param dbapi2: The DB-API 2 module used to create connections.
        This module must implement the :pep:`0249` specification.

    :param dict params: Connection parameters to pass directly to the
        :pep:`0249#connect` method when creating a new connection.

    :param float idlettl: The maximum time, in seconds, a connection can sit
        idle before it is discarded. If not specified, connections are retained
        indefinitely.

    :param int maxsize: The maximum number of connections to keep in
        the pool.

    :param bool block: Block until a connection is released back to the
        pool? This is only applicable when a `maxsize` value is specified. If
        `False`, a new connection will be created as needed, but the number of
        connections retained in the pool will never exceed `maxsize`.
    '''

    # Use __slots__ to minimize footprint.
    __slots__ = (
        '_block',
        '_connection_args',
        '_condition',
        '_dbapi2',
        '_idlettl',
        '_maxsize',
        '_nconnections',
        '_pool',
    )

    def __init__( # pylint: disable=too-many-arguments
            self,
            dbapi2,
            params,
            idlettl=None,
            maxsize=None,
            block=False,
    ):
        self._dbapi2 = dbapi2
        self._connection_args = params

        # The number of connections currently open.
        self._nconnections = 0

        self._condition = threading.Condition()

        self._maxsize = maxsize
        self._block = block
        self._idlettl = idlettl

        # Pool of connections. Ordered from least to most recently used.
        self._pool = []

    def __del__(self):
        self.finalize()

    @contextmanager
    def connection(self):
        '''
        Acquire a connection in a managed context for use with the Python
        `with` keyword.

        .. code-block:: python

            with pool.connection() as connection:
                # do stuff with connection

        '''
        connection = self.acquire()
        try:
            yield connection
        finally:
            self.release(connection)

    def acquire(self):
        '''
        Get a new connection from the pool.
        This will return an existing connection, if one is available in the
        pool, or create a new connection.

        .. warning:: If the pool was created with `maxsize` and `block=True`,
            this method may block until a connection is available in the pool.
        '''

        self._condition.acquire()
        try:
            # Wait for a connection if there is an upper bound to the pool.
            if self._maxsize is not None and self._block:
                while not self._pool and self._nconnections == self._maxsize:
                    self._condition.wait(timeout=None) # block indefinitely

            # Check the pool for a non-stale connection.
            while self._pool:
                pooledconn = self._pool.pop(0) # get least recently used connection
                if self._idlettl is not None and (pooledconn.released + self._idlettl) < time.time():
                    pooledconn.connection.close()
                    self._nconnections -= 1
                else:
                    return pooledconn.connection

            connection = self._dbapi2.connect(*(), **self._connection_args.copy())
            self._nconnections += 1

            return connection
        finally:
            self._condition.release()

    def release(self, connection):
        '''
        Return a connection back to the pool.

        Prior to release, :py:meth:`k_ctds.Connection.rollback()` is called to
        rollback any pending transaction.

        .. note:: This must be called once for every successful call to
            :py:meth:`.acquire()`.

        :param connection: The connection object returned by
            :py:meth:`.acquire()`.
        '''
        try:
            # Rollback the existing connection, closing on failure.
            connection.rollback()
        except self._dbapi2.Error:
            self._close(connection)
            return

        self._condition.acquire()
        try:
            if self._maxsize is None or self._maxsize > len(self._pool):
                self._pool.append(PooledConnection(connection, time.time()))
                self._condition.notify()
            else:
                self._close(connection)
        finally:
            self._condition.release()

    def finalize(self):
        '''
        Release all connections contained in the pool.

        .. note:: This should be called to cleanly shutdown the pool, i.e.
            on process exit.
        '''
        self._condition.acquire()
        try:
            if self._nconnections != len(self._pool):
                warnings.warn('finalize() called with unreleased connections', RuntimeWarning, 2)

            while self._pool:
                self._close(self._pool.pop().connection)
            self._nconnections = 0
        finally:
            self._condition.release()

    def _close(self, connection):
        assert self._nconnections > 0, \
            '.release() called multiple times for same connection'
        connection.close()
        self._nconnections -= 1



import re
import warnings

import k_ctds as ctds

from .base import TestExternalDatabase
from .compat import PY3, PY36, unicode_


class TestTdsParameter(TestExternalDatabase):

    def test___doc__(self):
        self.assertEqual(
            ctds.Parameter.__doc__,
            '''\
Parameter(value, output=False)

Explicitly define a parameter for :py:meth:`.callproc`,
:py:meth:`.execute`, or :py:meth:`.executemany`. This is necessary
to indicate whether a parameter is *SQL* `OUTPUT` or `INPUT/OUTPUT`
parameter.

:param object value: The parameter's value.
:param bool output: Is the parameter an output parameter.
'''
        )

    def test_parameter(self):
        param1 = ctds.Parameter(b'123', output=True)
        self.assertEqual(param1.value, b'123')
        self.assertTrue(isinstance(param1, ctds.Parameter))

        param2 = ctds.Parameter(b'123')
        self.assertEqual(param1.value, b'123')
        self.assertEqual(type(param1), type(param2))
        self.assertTrue(isinstance(param2, ctds.Parameter))

    def test___repr__(self):
        for parameter, expected in (
                (
                    ctds.Parameter(b'123', output=True),
                    "k_ctds.Parameter(b'123', output=True)" if PY3 else "k_ctds.Parameter('123', output=True)"
                ),
                (
                    ctds.Parameter(unicode_('123'), output=False),
                    "k_ctds.Parameter('123')" if PY3 else "k_ctds.Parameter(u'123')"
                ),
                (
                    ctds.Parameter(None),
                    "k_ctds.Parameter(None)"
                ),
                (
                    ctds.Parameter(ctds.SqlVarBinary(b'4321', size=10)),
                    "k_ctds.Parameter(k_ctds.SqlVarBinary(b'4321', size=10))"
                    if PY3 else
                    "k_ctds.Parameter(k_ctds.SqlVarBinary('4321', size=10))"
                )
        ):
            self.assertEqual(repr(parameter), expected)

    def _test__cmp__(self, __cmp__, expected, oper):
        cases = (
            (ctds.Parameter(b'1234'), ctds.Parameter(b'123')),
            (ctds.Parameter(b'123'), ctds.Parameter(b'123')),
            (ctds.Parameter(b'123'), ctds.Parameter(b'123', output=True)),
            (ctds.Parameter(b'123'), ctds.Parameter(b'1234')),
            (ctds.Parameter(b'123'), b'123'),
            (ctds.Parameter(b'123'), ctds.Parameter(123)),
            (ctds.Parameter(b'123'), unicode_('123')),
            (ctds.Parameter(b'123'), ctds.SqlBinary(None)),
            (ctds.Parameter(b'123'), 123),
            (ctds.Parameter(b'123'), None),
        )

        for index, args in enumerate(cases):
            operation = '[{0}]: {1} {2} {3}'.format(index, repr(args[0]), oper, repr(args[1]))
            if expected[index] == TypeError:
                try:
                    __cmp__(*args)
                except TypeError as ex:
                    regex = (
                        r"'{0}' not supported between instances of '[^']+' and '[^']+'".format(oper)
                        if not PY3 or PY36
                        else
                        r'unorderable types: \S+ {0} \S+'.format(oper)
                    )
                    self.assertTrue(re.match(regex, str(ex)), ex)
                else:
                    self.fail('{0} did not fail as expected'.format(operation)) # pragma: nocover
            else:
                self.assertEqual(__cmp__(*args), expected[index], operation)

    def test___cmp__eq(self):
        self._test__cmp__(
            lambda left, right: left == right,
            (
                False,
                True,
                True,
                False,
                True,
                False,
                not PY3,
                False,
                False,
                False,
            ),
            '=='
        )

    def test___cmp__ne(self):
        self._test__cmp__(
            lambda left, right: left != right,
            (
                True,
                False,
                False,
                True,
                False,
                True,
                PY3,
                True,
                True,
                True,
            ),
            '!='
        )

    def test___cmp__lt(self):
        self._test__cmp__(
            lambda left, right: left < right,
            (
                False,
                False,
                False,
                True,
                False,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
            ),
            '<'
        )

    def test___cmp__le(self):
        self._test__cmp__(
            lambda left, right: left <= right,
            (
                False,
                True,
                True,
                True,
                True,
                TypeError if PY3 else False,
                TypeError if PY3 else True,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
                TypeError if PY3 else False,
            ),
            '<='
        )

    def test___cmp__gt(self):
        self._test__cmp__(
            lambda left, right: left > right,
            (
                True,
                False,
                False,
                False,
                False,
                TypeError if PY3 else True,
                TypeError if PY3 else False,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
            ),
            '>'
        )

    def test___cmp__ge(self):
        self._test__cmp__(
            lambda left, right: left >= right,
            (
                True,
                True,
                True,
                False,
                True,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
                TypeError if PY3 else True,
            ),
            '>='
        )

    def test_typeerror(self):
        for case in (None, object(), 123, 'foobar'):
            self.assertRaises(TypeError, ctds.Parameter, case, b'123')

        self.assertRaises(TypeError, ctds.Parameter)
        self.assertRaises(TypeError, ctds.Parameter, output=False)

        for case in (None, object(), 123, 'foobar'):
            self.assertRaises(TypeError, ctds.Parameter, b'123', output=case)

    def test_reuse(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                for value in (
                        None,
                        123456,
                        unicode_('hello world'),
                        b'some bytes',
                ):
                    for output in (True, False):
                        parameter = ctds.Parameter(value, output=output)
                        for _ in range(0, 2):
                            # Ignore warnings generated due to output parameters
                            # used with result sets.
                            with warnings.catch_warnings(record=True):
                                cursor.execute(
                                    '''
                                    SELECT :0
                                    ''',
                                    (parameter,)
                                )
                                self.assertEqual(
                                    [tuple(row) for row in cursor.fetchall()],
                                    [(value,)]
                                )



import k_ctds as ctds

from .base import TestExternalDatabase
from .compat import long_

class TestCursorSpid(TestExternalDatabase):
    '''Unit tests related to the Cursor.spid property.
    '''

    def test___doc__(self):
        self.assertEqual(
            ctds.Cursor.spid.__doc__,
            '''\
Retrieve the SQL Server Session Process ID (SPID) for the connection or
:py:data:`None` if the connection is closed.

:rtype: int
'''
        )

    def test_closed(self):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.close()
            try:
                cursor.spid
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'cursor closed')
            else:
                self.fail('.spid did not fail as expected') # pragma: nocover

    def test_closed_connection(self):
        connection = self.connect()
        with connection.cursor() as cursor:
            connection.close()
            self.assertEqual(cursor.spid, None)

    def test_read(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                self.assertTrue(isinstance(cursor.spid, long_))
                cursor.execute('SELECT @@SPID;')
                spid = cursor.fetchone()[0]

                if self.freetds_version >= (1, 0, 0): # pragma: nocover
                    self.assertEqual(cursor.spid, spid)

    def test_write(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                try:
                    cursor.spid = 9
                except AttributeError:
                    pass
                else:
                    self.fail('.spid did not fail as expected') # pragma: nocover

                try:
                    cursor.spid = None
                except AttributeError:
                    pass
                else:
                    self.fail('.spid did not fail as expected') # pragma: nocover



import warnings

import k_ctds as ctds

from .base import TestExternalDatabase

class TestCursorFetchAll(TestExternalDatabase):
    '''Unit tests related to the Cursor.fetchall() method.
    '''
    def test___doc__(self):
        self.assertEqual(
            ctds.Cursor.fetchall.__doc__,
            '''\
fetchall()

Fetch all (remaining) rows of a query result, returning them as a
sequence of sequences.

:pep:`0249#fetchall`

:return: A sequence of result rows.
:rtype: k_ctds.RowList
'''
        )

    def test_closed(self):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.close()
            try:
                cursor.fetchall()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'cursor closed')
            else:
                self.fail('.fetchall() did not fail as expected') # pragma: nocover

    def test_closed_connection(self):
        connection = self.connect()
        with connection.cursor() as cursor:
            connection.close()
            try:
                cursor.fetchall()
            except ctds.InterfaceError as ex:
                self.assertEqual(str(ex), 'connection closed')
            else:
                self.fail('.fetchall() did not fail as expected') # pragma: nocover

    def test_premature(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                self.assertRaises(ctds.InterfaceError, cursor.fetchall)

    def test_fetchall(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT * FROM @{0};
                        SELECT i * 2 FROM @{0};
                    '''.format(self.test_fetchall.__name__)
                )
                self.assertEqual([tuple(row) for row in cursor.fetchall()], [(1,), (2,), (3,)])
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual([tuple(row) for row in cursor.fetchall()], [(2,), (4,), (6,)])
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), None)
                self.assertRaises(ctds.InterfaceError, cursor.fetchall)

    def test_empty_resultset(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT i FROM @{0} WHERE i < 0;
                    '''.format(self.test_empty_resultset.__name__)
                )
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), None)

    def test_multiple_resultsets(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                cursor.execute(
                    '''
                        DECLARE @{0} TABLE(i INT);
                        INSERT INTO @{0}(i) VALUES (1),(2),(3);
                        SELECT i FROM @{0} WHERE i < 0;
                        SELECT i AS j FROM @{0} WHERE i > 2;
                        SELECT i AS k FROM @{0} WHERE i > 3;
                        SELECT i AS ii FROM @{0};
                    '''.format(self.test_multiple_resultsets.__name__)
                )
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual([tuple(row) for row in cursor.fetchall()], [(3,)])
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual(list(cursor.fetchall()), [])
                self.assertEqual(cursor.nextset(), True)
                self.assertEqual([tuple(row) for row in cursor.fetchall()], [(1,), (2,), (3,)])
                self.assertEqual(cursor.nextset(), None)

class TestCursorFetchallWarningAsError(TestExternalDatabase):
    """Regression test for Cursor_fetchrows row-buffer leak (warning-as-error)."""

    def test_fetchrows_warning_as_error(self):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                sproc = self.test_fetchrows_warning_as_error.__name__
                with self.stored_procedure(
                    cursor,
                    sproc,
                    '''
                    AS
                        SELECT 1 AS col1, 'hello' AS col2;
                        RAISERROR (N'fetchrows leak test warning', 10, 1) WITH NOWAIT;
                    '''
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter('error', ctds.Warning)
                        cursor.callproc(sproc, ())

                        try:
                            cursor.fetchall()
                            self.fail('.fetchall() did not raise as expected')  # pragma: nocover
                        except ctds.Warning as ex:
                            self.assertIn('fetchrows leak test warning', str(ex))

